#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple, List

import numpy as np
import numba as nb

from scipy.special import comb

from functools import lru_cache

from .state import BaseFockState

from piquasso.api.instruction import Instruction
from piquasso.api.result import Result
from piquasso.api.exceptions import InvalidInstruction, InvalidParameter

from piquasso._math.fock import (
    cutoff_fock_space_dim,
    cutoff_fock_space_dim_array,
    operator_basis,
    get_fock_space_basis,
    nb_get_fock_space_basis,
)
from piquasso._math.indices import (
    get_index_in_fock_space,
    get_auxiliary_modes,
    get_index_in_fock_space_array,
    get_index_in_fock_subspace,
)


def attenuator(state: BaseFockState, instruction: Instruction, shots: int) -> Result:
    r"""
    Performs the deterministic loss or attenuation channel :math:`C` according to the
    equation

    .. math::
        C(| n \rangle \langle m |) =
            \sum_{k = 0}^{\min(n,m)} \tan(\theta)^{2k} \cos(\theta)^{n + m}
            \sqrt{ {n \choose k} {m \choose k} } | n - k \rangle \langle m - k |.
    """

    modes = instruction.modes

    if state._config.validate and len(modes) != 1:
        raise InvalidInstruction(
            f"The instruction should be specified for '{len(modes)}' "
            f"modes: instruction={instruction}"
        )

    mean_thermal_excitation = instruction._all_params["mean_thermal_excitation"]

    if state._config.validate and not np.isclose(mean_thermal_excitation, 0.0):
        raise InvalidParameter(
            "Non-zero mean thermal excitation is not supported in this backend. "
            f"mean_thermal_excitation={mean_thermal_excitation}"
        )

    theta = instruction._all_params["theta"]

    space = get_fock_space_basis(d=state.d, cutoff=state._config.cutoff)

    new_state = state._as_mixed()

    new_density_matrix = new_state._get_empty()

    for index, basis in operator_basis(space):
        coefficient = new_state._density_matrix[index]

        ket = np.copy(basis[0])
        bra = np.copy(basis[1])

        n = ket[modes]
        m = bra[modes]

        common_term = coefficient * np.cos(theta) ** (n + m)

        for k in range(min(n, m) + 1):
            ket[(modes,)] -= k
            bra[(modes,)] -= k

            current_ket_index = get_index_in_fock_space(ket)
            current_bra_index = get_index_in_fock_space(bra)

            ket[(modes,)] += k
            bra[(modes,)] += k

            current_index = (current_ket_index, current_bra_index)

            new_density_matrix[current_index] += common_term * (
                np.tan(theta) ** (2 * k) * np.sqrt(comb(n, k) * comb(m, k))
            )

    new_state._density_matrix = new_density_matrix

    return Result(state=new_state)


@nb.njit(cache=True)
def nb_calculate_state_index_matrix_list(d, cutoff, mode):
    auxiliary_subspace = nb_get_fock_space_basis(d=d - 1, cutoff=cutoff)

    auxiliary_modes = get_auxiliary_modes(d, (mode,))
    all_occupation_numbers = np.zeros(d, dtype=np.int32)

    indices = cutoff_fock_space_dim_array(cutoff=np.arange(cutoff + 1), d=d - 1)

    state_index_matrix_list = []

    for n in range(cutoff):
        limit = cutoff - n
        subspace_size = indices[n + 1] - indices[n]

        state_index_matrix = np.empty(shape=(limit, subspace_size), dtype=np.int32)

        for i, auxiliary_occupation_numbers in enumerate(
            auxiliary_subspace[indices[n] : indices[n + 1]]
        ):
            all_occupation_numbers[(auxiliary_modes,)] = auxiliary_occupation_numbers

            for j in range(limit):
                all_occupation_numbers[mode] = j
                index = get_index_in_fock_space(all_occupation_numbers)
                state_index_matrix[j, i] = index

        state_index_matrix_list.append(state_index_matrix)

    return state_index_matrix_list


calculate_state_index_matrix_list = lru_cache(maxsize=None)(
    nb_calculate_state_index_matrix_list
)


@nb.njit(cache=True)
def calculate_interferometer_helper_indices(d, cutoff):
    space = nb_get_fock_space_basis(d=d, cutoff=cutoff)

    basis = np.empty((space.shape[0], d), dtype=space.dtype)
    first_subpace_indices_space = np.empty(shape=space.shape[0], dtype=space.dtype)
    sqrt_first_occupation_numbers = np.empty(len(space), dtype=np.float64)

    first_nonzero_space_index = np.empty(shape=space.shape[0], dtype=space.dtype)

    sqrt_space = np.empty(shape=space.shape, dtype=np.float64)

    for i in range(len(space)):
        current_basis = space[i]
        sqrt_space[i] = np.sqrt(current_basis)
        found_first = False
        for j in range(d):
            current_basis[j] -= 1
            basis[i, j] = get_index_in_fock_subspace(current_basis)

            if not found_first and current_basis[j] >= 0:
                first_nonzero_space_index[i] = j
                first_subpace_indices_space[i] = basis[i, j]
                found_first = True
                sqrt_first_occupation_numbers[i] = sqrt_space[i, j]

            current_basis[j] += 1

    subspace_index_tensor = []

    first_subspace_index_tensor = []
    first_nonzero_index_tensor = []

    sqrt_occupation_numbers_tensor = []
    sqrt_first_occupation_numbers_tensor = []

    indices = cutoff_fock_space_dim_array(cutoff=np.arange(1, cutoff + 1), d=d)

    for n in range(2, cutoff):
        subspace_range = np.arange(indices[n - 1], indices[n])
        subspace_index_tensor.append(
            np.mod(basis[subspace_range], indices[n - 1] - indices[n - 2])
        )

        first_nonzero_index_tensor.append(first_nonzero_space_index[subspace_range])
        first_subspace_index_tensor.append(first_subpace_indices_space[subspace_range])

        sqrt_occupation_numbers_tensor.append(sqrt_space[subspace_range])
        sqrt_first_occupation_numbers_tensor.append(
            sqrt_first_occupation_numbers[subspace_range]
        )

    return (
        subspace_index_tensor,
        first_nonzero_index_tensor,
        first_subspace_index_tensor,
        sqrt_occupation_numbers_tensor,
        sqrt_first_occupation_numbers_tensor,
    )


@nb.njit(cache=True)
def nb_calculate_index_list_for_appling_interferometer(
    modes: Tuple[int, ...],
    d: int,
    cutoff: int,
) -> List[np.ndarray]:
    subspace = nb_get_fock_space_basis(d=len(modes), cutoff=cutoff)
    auxiliary_subspace = nb_get_fock_space_basis(d=d - len(modes), cutoff=cutoff)

    indices = cutoff_fock_space_dim_array(cutoff=np.arange(cutoff + 1), d=len(modes))
    auxiliary_indices = cutoff_fock_space_dim_array(
        cutoff=np.arange(cutoff + 1), d=d - len(modes)
    )
    auxiliary_modes = get_auxiliary_modes(d, modes)

    all_occupation_numbers = np.zeros(d, dtype=np.int32)

    index_list = []

    for n in range(cutoff):
        size = indices[n + 1] - indices[n]
        n_particle_subspace = subspace[indices[n] : indices[n + 1]]
        auxiliary_size = auxiliary_indices[cutoff - n]
        state_index_matrix = np.empty(shape=(size, auxiliary_size), dtype=np.int32)
        for idx1, auxiliary_occupation_numbers in enumerate(
            auxiliary_subspace[:auxiliary_size]
        ):
            for idx, mode in enumerate(auxiliary_modes):
                all_occupation_numbers[mode] = auxiliary_occupation_numbers[idx]

            for idx2, column_vector_on_subspace in enumerate(n_particle_subspace):
                for idx, mode in enumerate(modes):
                    all_occupation_numbers[mode] = column_vector_on_subspace[idx]

                column_index = get_index_in_fock_space(all_occupation_numbers)
                state_index_matrix[idx2, idx1] = column_index

        index_list.append(state_index_matrix)

    return index_list


calculate_index_list_for_appling_interferometer = lru_cache(maxsize=None)(
    nb_calculate_index_list_for_appling_interferometer
)


def get_projection_operator_indices(d, cutoff, modes, basis_vector):
    new_cutoff = cutoff - np.sum(basis_vector)

    boxes = d - len(modes)

    card = cutoff_fock_space_dim(cutoff=new_cutoff, d=boxes)

    basis = np.empty(shape=(card, d), dtype=int)

    basis[:, modes] = basis_vector

    if len(modes) < d:
        auxiliary_modes = get_auxiliary_modes(d, modes)

        auxiliary_subspace = get_fock_space_basis(d=boxes, cutoff=new_cutoff)

        basis[:, auxiliary_modes] = auxiliary_subspace

    return get_index_in_fock_space_array(basis)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np

import piquasso as pq

for_all_connectors = pytest.mark.parametrize(
    "connector", [pq.NumpyConnector(), pq.JaxConnector()]
)


@for_all_connectors
def test_PureFockState_d(connector):
    d = 2

    with pq.Program() as program:
        pq.Q(0, 1) | pq.StateVector([1, 1])

    simulator = pq.fermionic.PureFockSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    assert state.d == d


@for_all_connectors
def test_PureFockState_fock_probabilities(connector):
    theta = np.pi / 5

    U = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])

    with pq.Program() as program:
        pq.Q(0, 1) | pq.StateVector([0, 1])
        pq.Q(0, 1) | pq.Interferometer(U)

    simulator = pq.fermionic.PureFockSimulator(d=2, connector=connector)

    state = simulator.execute(program).state

    assert np.allclose(
        state.fock_probabilities, [0, np.sin(theta) ** 2, np.cos(theta) ** 2, 0]
    )


@for_all_connectors
def test_PureFockState_norm(connector):
    theta = np.pi / 5

    U = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])

    with pq.Program() as program:
        pq.Q(0, 1) | pq.StateVector([0, 1]) / np.sqrt(2)
        pq.Q(0, 1) | pq.Interferometer(U)

    simulator = pq.fermionic.PureFockSimulator(d=2, connector=connector)

    state = simulator.execute(program).state

    assert np.allclose(state.norm, 1 / 2)


@for_all_connectors
def test_PureFockState_get_particle_detection_probability(connector):
    theta = np.pi / 5

    U = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])

    with pq.Program() as program:
        pq.Q(0, 1) | pq.StateVector([0, 1])
        pq.Q(0, 1) | pq.Interferometer(U)

    simulator = pq.fermionic.PureFockSimulator(d=2, connector=connector)

    state = simulator.execute(program).state

    assert np.allclose(
        state.get_particle_detection_probability([0, 1]), np.cos(theta) ** 2
    )


@for_all_connectors
def test_PureFockState_validate_unnormalized(connector):
    theta = np.pi / 5

    U = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])

    with pq.Program() as program:
        pq.Q(0, 1) | pq.StateVector([0, 1]) / np.sqrt(2)
        pq.Q(0, 1) | pq.Interferometer(U)

    simulator = pq.fermionic.PureFockSimulator(d=2, connector=connector)

    state = simulator.execute(program).state

    with pytest.raises(pq.api.exceptions.InvalidState) as error:
        state.validate()

    assert "The sum of probabilities is" in error.value.args[0]


@for_all_connectors
def test_PureFockState_validate_with_validation_turned_off(connector):
    theta = np.pi / 5

    U = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])

    with pq.Program() as program:
        pq.Q(0, 1) | pq.StateVector([0, 1]) / np.sqrt(2)
        pq.Q(0, 1) | pq.Interferometer(U)

    simulator = pq.fermionic.PureFockSimulator(
        d=2, connector=connector, config=pq.Config(validate=False)
    )

    state = simulator.execute(program).state

    state.validate()


@for_all_connectors
def test_PureFockState_eq_with_itself(connector):
    theta = np.pi / 5

    U = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])

    with pq.Program() as program:
        pq.Q(0, 1) | pq.StateVector([0, 1]) / np.sqrt(2)
        pq.Q(0, 1) | pq.Interferometer(U)

    simulator = pq.fermionic.PureFockSimulator(
        d=2, connector=connector, config=pq.Config(validate=False)
    )

    state = simulator.execute(program).state

    assert state == state


@for_all_connectors
def test_PureFockState_eq_different_type(connector):
    theta = np.pi / 5

    U = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])

    with pq.Program() as program:
        pq.Q(0, 1) | pq.StateVector([0, 1]) / np.sqrt(2)
        pq.Q(0, 1) | pq.Interferometer(U)

    simulator = pq.fermionic.PureFockSimulator(
        d=2, connector=connector, config=pq.Config(validate=False)
    )

    state = simulator.execute(program).state

    some_other_object = object()

    assert state != some_other_object


@for_all_connectors
def test_PureFockState_fock_probabilities_map(connector):
    theta = np.pi / 5

    U = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])

    with pq.Program() as program:
        pq.Q(0, 1) | pq.StateVector([0, 1])
        pq.Q(0, 1) | pq.Interferometer(U)

    simulator = pq.fermionic.PureFockSimulator(d=2, connector=connector)

    state = simulator.execute(program).state

    expected = {
        (0, 0): 0,
        (1, 0): np.sin(theta) ** 2,
        (0, 1): np.cos(theta) ** 2,
        (1, 1): 0,
    }

    for occupation_number, probability in expected.items():
        assert np.isclose(state.fock_probabilities_map[occupation_number], probability)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import tensorflow as tf

import strawberryfields as sf
import piquasso as pq


def pq_cvnn_gradient(weights, d, cutoff, layer_count):
    simulator = pq.PureFockSimulator(
        d=d,
        config=pq.Config(cutoff=cutoff, normalize=True),
        connector=pq.TensorflowConnector(),
    )

    with tf.GradientTape() as tape:
        with pq.Program() as program:
            pq.Q(0) | pq.Vacuum()
            for i in range(layer_count):
                pq.Q(0) | pq.Phaseshifter(weights[i, 0])
                pq.Q(0) | pq.Squeezing(weights[i, 1], weights[i, 2])
                pq.Q(0) | pq.Phaseshifter(weights[i, 3])
                pq.Q(0) | pq.Displacement(weights[i, 4], weights[i, 5])
                pq.Q(0) | pq.Kerr(weights[i, 6])

        state = simulator.execute(program).state
        ket = state.state_vector

    return state, ket, tape.gradient(ket, weights)


def sf_cvnn_gradient(weights, d, cutoff, layer_count):
    eng = sf.Engine(backend="tf", backend_options={"cutoff_dim": cutoff})
    prog = sf.Program(d)

    with tf.GradientTape() as tape:
        with prog.context as q:
            for i in range(layer_count):
                sf.ops.Rgate(weights[i, 0]) | q[0]
                sf.ops.Sgate(weights[i, 1], weights[i, 2]) | q[0]
                sf.ops.Rgate(weights[i, 3]) | q[0]
                sf.ops.Dgate(weights[i, 4], weights[i, 5]) | q[0]
                sf.ops.Kgate(weights[i, 6]) | q[0]

        state = eng.run(prog).state
        ket = state.ket()

    return state, ket, tape.gradient(ket, weights)


def cvnn_comparison():
    cutoff = 12
    d = 1
    layer_count = 3
    tf.random.set_seed(42)
    np.random.seed(42)

    passive_sd = 0.1
    active_sd = 0.001

    sq_r = tf.random.normal(shape=[layer_count], stddev=active_sd)
    sq_phi = tf.random.normal(shape=[layer_count], stddev=passive_sd)

    d_r = tf.random.normal(shape=[layer_count], stddev=active_sd)
    d_phi = tf.random.normal(shape=[layer_count], stddev=passive_sd)

    r1 = tf.random.normal(shape=[layer_count], stddev=passive_sd)
    r2 = tf.random.normal(shape=[layer_count], stddev=passive_sd)

    kappa = tf.random.normal(shape=[layer_count], stddev=active_sd)

    weights = tf.convert_to_tensor(
        [r1, sq_r, sq_phi, r2, d_r, d_phi, kappa], dtype=np.float64
    )
    weights = tf.Variable(tf.transpose(weights))
    pq_state, pq_ket, pq_grad = pq_cvnn_gradient(weights, d, cutoff, layer_count)
    sf_state, sf_ket, sf_grad = sf_cvnn_gradient(weights, d, cutoff, layer_count)
    assert np.allclose(pq_ket, sf_ket)
    assert np.allclose(np.asarray(pq_grad), np.asarray(sf_grad))


if __name__ == "__main__":
    cvnn_comparison()



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import strawberryfields as sf
from strawberryfields import ops
import piquasso as pq
import time
import numpy as np
import json

from piquasso import cvqnn

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

TIMESTAMP = time.strftime("%Y%m%d-%H%M%S")

PQ_FILE_NAME = f"pq_results_{TIMESTAMP}.json"
SF_FILE_NAME = f"sf_results_{TIMESTAMP}.json"

PQ_CUTOFF_RANGE = range(3, 16 + 1)
SF_CUTOFF_RANGE = range(3, 6 + 1)

ACTIVE_VAR = 0.1
PASSIVE_VAR = 1.0

WARMUP = 10
ITERATIONS = 100
d = 10
number_of_layers = 4


def pq_benchmark():
    print("PIQUASSO_START")

    benchmark_json_list = {
        "d": d,
        "number_of_layers": number_of_layers,
        "benchmarks": [],
    }

    for cutoff in PQ_CUTOFF_RANGE:
        print(f"cutoff={cutoff}")

        simulator = pq.PureFockSimulator(d=d, config=pq.Config(cutoff=cutoff))

        exec_time_list = []
        probability_lost_list = []

        for _ in range(ITERATIONS + WARMUP):
            weights = cvqnn.generate_random_cvqnn_weights(
                number_of_layers, d, active_var=ACTIVE_VAR, passive_var=PASSIVE_VAR
            )
            program = cvqnn.create_program(weights)

            start_time = time.time()
            state = simulator.execute(program).state
            exec_time = time.time() - start_time

            exec_time_list.append(exec_time)

            probability_lost_list.append(1.0 - state.norm)

        average_exec_time = np.average(exec_time_list[WARMUP:])
        average_probability_lost = np.average(probability_lost_list[WARMUP:])

        benchmark_values = {
            "cutoff": cutoff,
            "probability_lost": average_probability_lost,
            "exec_time": average_exec_time,
            "iterations": ITERATIONS,
        }

        benchmark_json_list["benchmarks"].append(benchmark_values)

        with open(PQ_FILE_NAME, "w") as file:
            json.dump(benchmark_json_list, file, indent=6)


def sf_benchmark():
    print("STRAWBERRYFIELDS_START")
    """
    This code has been copied from the following website with minor modifications:
    https://strawberryfields.ai/photonics/demos/run_quantum_neural_network.html
    """

    def interferometer(params, q):
        N = len(q)
        theta = params[: N * (N - 1) // 2]
        phi = params[N * (N - 1) // 2 : N * (N - 1)]
        rphi = params[-N + 1 :]

        if N == 1:
            ops.Rgate(rphi[0]) | q[0]
            return

        n = 0

        for i in range(N):
            for j, (q1, q2) in enumerate(zip(q[:-1], q[1:])):
                if (i + j) % 2 != 1:
                    ops.BSgate(theta[n], phi[n]) | (q1, q2)
                    n += 1

        for i in range(max(1, N - 1)):
            ops.Rgate(rphi[i]) | q[i]

    def layer(params, q):
        N = len(q)
        M = int(N * (N - 1)) + max(1, N - 1)

        int1 = params[:M]
        s = params[M : M + N]
        int2 = params[M + N : 2 * M + N]
        dr = params[2 * M + N : 2 * M + 2 * N]
        dp = params[2 * M + 2 * N : 2 * M + 3 * N]
        k = params[2 * M + 3 * N : 2 * M + 4 * N]

        # begin layer
        interferometer(int1, q)

        for i in range(N):
            ops.Sgate(s[i]) | q[i]

        interferometer(int2, q)

        for i in range(N):
            ops.Dgate(dr[i], dp[i]) | q[i]
            ops.Kgate(k[i]) | q[i]

    benchmark_json_list = {
        "d": d,
        "number_of_layers": number_of_layers,
        "benchmarks": [],
    }

    for cutoff in SF_CUTOFF_RANGE:
        print(f"cutoff={cutoff}")

        exec_time_list = []
        probability_lost_list = []

        for _ in range(ITERATIONS + WARMUP):
            eng = sf.Engine(backend="fock", backend_options={"cutoff_dim": cutoff})
            qnn = sf.Program(d)
            weights = cvqnn.generate_random_cvqnn_weights(
                number_of_layers, d, active_var=ACTIVE_VAR, passive_var=PASSIVE_VAR
            )

            with qnn.context as q:
                for k in range(number_of_layers):
                    layer(weights[k], q)

            start_time = time.time()
            state = eng.run(qnn).state
            exec_time = time.time() - start_time
            exec_time_list.append(exec_time)

            ket = state.ket()
            flatten_ket = np.reshape(ket, (-1))
            probability_lost = 1.0 - np.sqrt(
                np.real(np.dot(flatten_ket, np.conj(flatten_ket)))
            )

            probability_lost_list.append(probability_lost)

        average_exec_time = np.average(exec_time_list[WARMUP:])
        average_probability_lost = np.average(probability_lost_list[WARMUP:])

        benchmark_values = {
            "cutoff": cutoff,
            "probability_lost": average_probability_lost,
            "exec_time": average_exec_time,
            "iterations": ITERATIONS,
        }

        benchmark_json_list["benchmarks"].append(benchmark_values)

        with open(SF_FILE_NAME, "w") as file:
            json.dump(benchmark_json_list, file, indent=6)


if __name__ == "__main__":
    print(f"ITERATIONS={ITERATIONS}, d={d}, number_of_layers={number_of_layers}")

    pq_benchmark()
    sf_benchmark()



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import numba as nb

from piquasso.api.exceptions import InvalidParameter

from piquasso._math.linalg import is_selfadjoint, is_skew_symmetric
from piquasso._math.combinatorics import comb


def validate_fermionic_gaussian_hamiltonian(H):
    r"""
    Checks if `H` is a fermionic quadratic Hamiltonian in the Dirac representation.

    More concretely, it validates if :math:`H \in \mathbb{C}^{2d \times 2d}` is a
    self-adjoint matrix of the form

    .. math::
        H = \begin{bmatrix}
            A & -\overline{B} \\
            B & -\overline{A}
        \end{bmatrix}.

    where :math:`A^\dagger = A` and :math:`B^T = - B`.
    """
    d = len(H) // 2

    A = H[d:, d:]
    B = H[:d, d:]

    H11 = H[:d, :d]
    H21 = H[d:, :d]

    if (
        not is_selfadjoint(A)
        or not is_skew_symmetric(B)
        or not np.allclose(-A.conj(), H11)
        or not np.allclose(-B.conj(), H21)
    ):
        raise InvalidParameter("Invalid Hamiltonian specified.")


def tensor_product(ops):
    if len(ops) == 1:
        return ops[0]

    return np.kron(ops[0], tensor_product(ops[1:]))


def _embed_f(op, index, d):
    r"""Embeds the single-mode Dirac operators into the Fock space.

    This function returns with

    .. math::
        f_i &= Z^{\otimes (i-1)} \otimes f \otimes, I^{d-i} \\\\
        f_i^\dagger &= Z^{\otimes (i-1)} \otimes f^\dagger \otimes I^{d-i},

    where :math:`f` and :math:`f^\dagger` are the single-mode annihilation and creation
    operators, respectively.
    """

    Z = np.diag([1, -1])
    I = np.identity(2)

    ops = [Z] * index

    ops.append(op)
    ops += [I] * (d - index - 1)

    return tensor_product(ops)


def _fdag():
    return np.array(
        [
            [0, 0],
            [1, 0],
        ],
        dtype=complex,
    )


def _f():
    return np.array(
        [
            [0, 1],
            [0, 0],
        ],
        dtype=complex,
    )


def _get_fs_fdags(d):
    fs = []
    fdags = []

    for i in range(d):
        fs.append(_embed_f(_f(), i, d))
        fdags.append(_embed_f(_fdag(), i, d))

    return fs, fdags


def get_fermionic_hamiltonian(H, connector):
    r"""Calculates the fermionic Hamiltonian.

    This function returns

    .. math::
        \hat{H} = \mathbf{f} H \mathbf{f}^\dagger,

    where :math:`H \in \mathbb{C}^{2d \times 2d}` is a self-adjoint matrix of the form

    .. math::
        H = \begin{bmatrix}
            A & -\overline{B} \\
            B & -\overline{A}
        \end{bmatrix}.

    where :math:`A^\dagger = A` and :math:`B^T = - B`.
    """

    d = len(H) // 2

    bigH = connector.np.zeros(shape=(2**d, 2**d), dtype=complex)

    fs, fdags = _get_fs_fdags(d)

    f = fs + fdags

    for i in range(2 * d):
        for j in range(2 * d):
            bigH += H[i, j] * f[i] @ f[j].T

    return bigH


def get_omega(d: int) -> np.ndarray:
    """
    Basis transformation from real to complex basis.
    """

    identity = np.identity(d)

    return np.block(
        [
            [identity, identity],
            [1j * identity, -1j * identity],
        ]
    ) / np.sqrt(2)


@nb.njit(cache=True)
def get_cutoff_fock_space_dimension(d, cutoff):
    sum_ = 0
    for k in range(cutoff):
        sum_ += get_fock_subspace_dimension(d, k)

    return sum_


@nb.njit(cache=True)
def cutoff_fock_space_dim_array(cutoff, d):
    ret = np.empty(cutoff.shape, dtype=np.int64)

    for i in range(len(cutoff)):
        ret[i] = get_cutoff_fock_space_dimension(d, cutoff[i])

    return ret


@nb.njit(cache=True)
def get_fock_subspace_dimension(d, k):
    return comb(d, k)


@nb.njit(cache=True)
def get_fock_space_index(occupation_numbers):
    return _get_fock_space_index_first_quantized(
        _to_first_quantized(occupation_numbers), len(occupation_numbers)
    )


@nb.njit(cache=True)
def get_fock_subspace_index(occupation_numbers):
    return get_fock_subspace_index_first_quantized(
        _to_first_quantized(occupation_numbers), len(occupation_numbers)
    )


@nb.njit(cache=True)
def get_fock_subspace_index_first_quantized(first_quantized, d):
    n = len(first_quantized)

    if n == 0:
        return 0

    sum_ = comb(d, n) - 1

    for i in range(n):
        sum_ -= comb(d - first_quantized[i] - 1, n - i)

    return sum_


@nb.njit(cache=True)
def _get_fock_space_index_first_quantized(first_quantized, d):
    n = len(first_quantized)

    return get_cutoff_fock_space_dimension(
        d, n
    ) + get_fock_subspace_index_first_quantized(first_quantized, d)


@nb.njit(cache=True)
def _to_first_quantized(occupation_numbers):
    n = sum(occupation_numbers)
    first_quantized = np.zeros(n, dtype=nb.int64)

    j = 0
    for i in range(len(occupation_numbers)):
        if occupation_numbers[i] == 1:
            first_quantized[j] = i
            j += 1

    return first_quantized


@nb.njit(cache=True)
def next_first_quantized(first_quantized, d):
    l = len(first_quantized)

    for i in range(l):
        if first_quantized[l - i - 1] < d - i - 1:
            first_quantized[l - i - 1] += 1
            for k in range(l - i, l):
                first_quantized[k] = first_quantized[l - i - 1] + k - l + i + 1

            return first_quantized

    next_first_quantized = np.empty(l + 1, dtype=first_quantized.dtype)

    for i in range(len(next_first_quantized)):
        next_first_quantized[i] = i

    return next_first_quantized


@nb.njit(cache=True)
def next_second_quantized(second_quantized):
    d = len(second_quantized)

    return _to_second_quantized(
        next_first_quantized(_to_first_quantized(second_quantized), d), d
    )


@nb.njit(cache=True)
def _to_second_quantized(first_quantized, d):
    ret = np.zeros(shape=d, dtype=nb.int64)

    for i in range(len(first_quantized)):
        ret[first_quantized[i]] = 1

    return ret


def binary_to_fock_indices(d):
    size = get_cutoff_fock_space_dimension(d, d + 1)

    power_array = np.empty(d, dtype=int)

    for i in range(d):
        power_array[i] = 2 ** (d - 1 - i)

    indices = np.empty(size, dtype=int)

    indices[0] = 0

    index = np.array([], dtype=int)

    for i in range(1, size):
        index = next_first_quantized(index, d)
        second_quantized = _to_second_quantized(index, d)

        indices[i] = second_quantized @ power_array

    return indices


@nb.njit(cache=True)
def get_fock_space_basis(d, cutoff):
    size = get_cutoff_fock_space_dimension(d, cutoff)

    basis = np.empty((size, d), dtype=np.int64)

    basis[0] = np.zeros(d, dtype=np.int64)

    for i in range(1, size):
        basis[i] = next_second_quantized(basis[i - 1])

    return basis



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable, Union, Tuple

import numpy as np


def is_natural(number: Union[int, float]) -> bool:
    return bool(np.isclose(number % 1, 0.0) and round(number) >= 0)


def all_natural(array: Iterable) -> bool:
    return all(is_natural(number) for number in array)


def is_zero_or_one(number: Union[int, float]) -> bool:
    return bool(np.isclose(number, 0.0) or np.isclose(number, 1.0))


def all_zero_or_one(array: Iterable) -> bool:
    return all(is_zero_or_one(number) for number in array)


def all_real_and_positive(vector: Iterable) -> bool:
    return all(element >= 0.0 or np.isclose(element, 0.0) for element in vector)


def all_in_interval(vector: Iterable, lower: float, upper: float) -> bool:
    return all(
        (element >= lower or np.isclose(element, lower))
        and (element <= upper or np.isclose(element, upper))
        for element in vector
    )


def are_modes_consecutive(modes: Tuple[int, ...]) -> bool:
    expected = np.arange(modes[0], modes[-1] + 1)

    return len(modes) == len(expected) and bool(np.all(modes == expected))



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

import piquasso as pq


def test_Loss_uniform():
    d = 5
    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 1, 1, 0, 0])

        for i in range(5):
            pq.Q(i) | pq.Loss(0.9)

    simulator = pq.SamplingSimulator(d=d)
    state = simulator.execute(program, shots=1).state

    assert state.is_lossy

    singular_values = np.linalg.svd(state.interferometer)[1]

    assert np.allclose(singular_values, [0.9, 0.9, 0.9, 0.9, 0.9])


def test_Loss_non_uniform():
    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 1, 1, 0, 0])

        pq.Q(0) | pq.Loss(transmissivity=0.4)
        pq.Q(1) | pq.Loss(transmissivity=0.5)

    simulator = pq.SamplingSimulator(d=5)
    state = simulator.execute(program, shots=1).state

    assert state.is_lossy

    singular_values = np.linalg.svd(state.interferometer)[1]

    assert len(singular_values[np.isclose(singular_values, 0.4)]) == 1
    assert len(singular_values[np.isclose(singular_values, 0.5)]) == 1
    assert len(singular_values[np.isclose(singular_values, 1.0)]) == 3



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np

import piquasso as pq


def test_PureFockState_reduced():
    with pq.Program() as program:
        pq.Q() | pq.StateVector([0, 1]) / 2

        pq.Q() | pq.StateVector([0, 2]) / 2
        pq.Q() | pq.StateVector([2, 0]) / np.sqrt(2)

    simulator = pq.PureFockSimulator(d=2)
    state = simulator.execute(program).state

    with pq.Program() as reduced_program:
        pq.Q() | pq.DensityMatrix(ket=(0,), bra=(0,)) / 2

        pq.Q() | pq.DensityMatrix(ket=(1,), bra=(1,)) / 4
        pq.Q() | pq.DensityMatrix(ket=(2,), bra=(2,)) / 4

        pq.Q() | pq.DensityMatrix(ket=(1,), bra=(2,)) / 4
        pq.Q() | pq.DensityMatrix(ket=(2,), bra=(1,)) / 4

    reduced_program_simulator = pq.FockSimulator(d=1)
    reduced_program_state = reduced_program_simulator.execute(reduced_program).state

    expected_reduced_state = reduced_program_state

    reduced_state = state.reduced(modes=(1,))

    assert expected_reduced_state == reduced_state


def test_PureFockState_reduced_preserves_Config():
    with pq.Program() as program:
        pq.Q() | pq.StateVector([0, 1]) / 2

        pq.Q() | pq.StateVector([0, 2]) / 2
        pq.Q() | pq.StateVector([2, 0]) / np.sqrt(2)

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=10))
    state = simulator.execute(program).state

    assert state._config.cutoff == 10


def test_PureFockState_fock_amplitudes_map():
    theta = 0.3
    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 0]) / np.sqrt(2)
        pq.Q() | pq.StateVector([0, 1]) / 2

        pq.Q() | pq.StateVector([0, 2]) / 2

        pq.Q(0) | pq.Phaseshifter(theta)

    simulator = pq.PureFockSimulator(d=2)
    state = simulator.execute(program).state

    expected_fock_amplitudes = {
        (0, 1): 0.5 + 0.0j,
        (0, 2): 0.5 + 0.0j,
        (1, 0): 1 / np.sqrt(2) * np.exp(1j * theta),
    }

    actual_fock_amplitudes = state.fock_amplitudes_map

    assert len(actual_fock_amplitudes.items()) == len(expected_fock_amplitudes.items())

    for occupation_number, expected_ampl in expected_fock_amplitudes.items():
        assert np.isclose(
            actual_fock_amplitudes[occupation_number],
            expected_ampl,
        )

    # Try some Fock states that are expected to have zero amplitudes
    assert np.isclose(actual_fock_amplitudes[(0, 0)], 0.0)
    assert np.isclose(actual_fock_amplitudes[(2, 1)], 0.0)
    assert np.isclose(actual_fock_amplitudes[(122, 35)], 0.0)


def test_PureFockState_fock_probabilities_map():
    with pq.Program() as program:
        pq.Q() | pq.StateVector([0, 1]) / 2

        pq.Q() | pq.StateVector([0, 2]) / 2
        pq.Q() | pq.StateVector([2, 0]) / np.sqrt(2)

    simulator = pq.PureFockSimulator(d=2)
    state = simulator.execute(program).state

    expected_fock_probabilities = {
        (0, 0): 0.0,
        (0, 1): 0.25,
        (1, 0): 0.0,
        (0, 2): 0.25,
        (1, 1): 0.0,
        (2, 0): 0.5,
        (0, 3): 0.0,
        (1, 2): 0.0,
        (2, 1): 0.0,
        (3, 0): 0.0,
    }

    actual_fock_probabilities = state.fock_probabilities_map

    for occupation_number, expected_probability in expected_fock_probabilities.items():
        assert np.isclose(
            actual_fock_probabilities[occupation_number],
            expected_probability,
        )


def test_PureFockState_get_marginal_fock_probabilities():
    with pq.Program() as program:
        pq.Q() | pq.StateVector([0, 1]) / 2

        pq.Q() | pq.StateVector([0, 2]) / 2
        pq.Q() | pq.StateVector([2, 0]) / np.sqrt(2)

    simulator = pq.PureFockSimulator(d=2)
    state = simulator.execute(program).state

    modes = (1,)
    expected_probabilities = np.array([0.5, 0.25, 0.25, 0.0])
    actual_probabilities = state.get_marginal_fock_probabilities(modes)

    assert np.allclose(actual_probabilities, expected_probabilities)


def test_PureFockState_quadratures_mean_variance():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=0.2, phi=0)
        pq.Q(1) | pq.Squeezing(r=0.1, phi=1.0)
        pq.Q(2) | pq.Displacement(r=0.2, phi=np.pi / 2)

    simulator = pq.PureFockSimulator(d=3, config=pq.Config(cutoff=8, hbar=2))
    result = simulator.execute(program)

    mean_1, var_1 = result.state.quadratures_mean_variance(modes=(0,))
    mean_2, var_2 = result.state.quadratures_mean_variance(modes=(1,))
    mean_3, var_3 = result.state.quadratures_mean_variance(modes=(0,), phi=np.pi / 2)
    mean_4, var_4 = result.state.quadratures_mean_variance(modes=(2,), phi=np.pi / 2)

    assert np.isclose(mean_1, 0.4, rtol=0.00001)
    assert mean_2 == 0.0
    assert np.isclose(mean_3, 0.0)
    assert np.isclose(mean_4, 0.4, rtol=0.00001)
    assert np.isclose(var_1, 1.0, rtol=0.00001)
    assert np.isclose(var_2, 0.9112844, rtol=0.00001)
    assert np.isclose(var_3, 1, rtol=0.00001)
    assert np.isclose(var_4, 1, rtol=0.00001)


def test_PureFockState_mean_position():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=0.2, phi=0)

    simulator = pq.PureFockSimulator(d=1, config=pq.Config(cutoff=8))
    state = simulator.execute(program).state

    assert np.isclose(state.mean_position(mode=0), 0.4)


def test_mean_position():
    d = 1
    cutoff = 7

    alpha_ = 0.02

    with pq.Program() as program:
        pq.Q(all) | pq.Vacuum()

        pq.Q(all) | pq.Displacement(r=alpha_)

    config = pq.Config(cutoff=cutoff)

    simulator = pq.PureFockSimulator(d=d, config=config)

    state = simulator.execute(program).state
    mean = state.mean_position(mode=0)

    assert np.allclose(mean, np.sqrt(2 * config.hbar) * alpha_)


def test_normalize_if_disabled_in_Config():
    d = 1
    cutoff = 3

    alpha_ = 1.0

    with pq.Program() as program:
        pq.Q(all) | pq.Vacuum()

        pq.Q(all) | pq.Displacement(r=alpha_)

    config = pq.Config(cutoff=cutoff)

    simulator = pq.PureFockSimulator(d=d, config=config)

    state = simulator.execute(program).state
    norm = state.norm

    assert not np.isclose(norm, 1.0)


@pytest.mark.parametrize("connector", (pq.NumpyConnector(), pq.JaxConnector()))
def test_PureFockState_get_tensor_representation(connector):
    d = 2
    cutoff = 3

    with pq.Program() as program:
        pq.Q() | pq.StateVector([0, 1]) / 2

        pq.Q() | pq.StateVector([0, 2]) / 2
        pq.Q() | pq.StateVector([2, 0]) / np.sqrt(2)

    simulator = pq.PureFockSimulator(
        d=d, config=pq.Config(cutoff=cutoff), connector=connector
    )
    state = simulator.execute(program).state

    state_tensor = state.get_tensor_representation()

    assert state_tensor.shape == (3,) * 2

    assert np.allclose(
        state_tensor,
        np.array(
            [
                [0.0 + 0.0j, 0.5 + 0.0j, 0.5 + 0.0j],
                [0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j],
                [0.70710678 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j],
            ]
        ),
    )


def test_PureFockState_get_purity():
    d = 2
    cutoff = 3

    with pq.Program() as program:
        pq.Q() | pq.StateVector([0, 1]) / 2

        pq.Q() | pq.StateVector([0, 2]) / 2
        pq.Q() | pq.StateVector([2, 0]) / np.sqrt(2)

    simulator = pq.PureFockSimulator(d=d, config=pq.Config(cutoff=cutoff))
    state = simulator.execute(program).state

    purity = state.get_purity()

    assert np.isclose(purity, 1.0)


@pytest.mark.monkey
def test_PureFockState_get_particle_detection_probability_on_modes():
    d = 2
    cutoff = 3

    state_vector_length = 6

    coeffs = np.random.uniform(size=state_vector_length) + 1j * np.random.uniform(
        size=state_vector_length
    )

    coeffs /= np.linalg.norm(coeffs)

    with pq.Program() as program:
        pq.Q() | pq.StateVector([0, 0]) * coeffs[0]
        pq.Q() | pq.StateVector([0, 1]) * coeffs[1]
        pq.Q() | pq.StateVector([1, 0]) * coeffs[2]
        pq.Q() | pq.StateVector([0, 2]) * coeffs[3]
        pq.Q() | pq.StateVector([1, 1]) * coeffs[4]
        pq.Q() | pq.StateVector([2, 0]) * coeffs[5]

    simulator = pq.PureFockSimulator(d=d, config=pq.Config(cutoff=cutoff))
    state = simulator.execute(program).state

    state.validate()

    probability_of_0_on_mode_1 = state.get_particle_detection_probability_on_modes(
        occupation_numbers=(0,), modes=(1,)
    )

    assert np.isclose(
        probability_of_0_on_mode_1,
        np.abs(coeffs[0]) ** 2 + np.abs(coeffs[2]) ** 2 + np.abs(coeffs[5]) ** 2,
    )


def test_PureFockState_displaced_squeezed_state_mean_photon_number():
    d = 2
    cutoff = 10

    r_S = 0.2
    r_D = 0.3

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(r_S, phi=np.pi / 5)
        pq.Q(0) | pq.Displacement(r_D, phi=np.pi / 3)

        pq.Q(0, 1) | pq.Beamsplitter()

    simulator = pq.PureFockSimulator(d=d, config=pq.Config(cutoff=cutoff))
    state = simulator.execute(program).state

    mean = state.mean_photon_number()

    expected_mean = np.sinh(r_S) ** 2 + r_D**2

    assert np.isclose(mean, expected_mean)


def test_PureFockState_displaced_state_variance_photon_number():
    d = 2
    cutoff = 10

    r = 0.3

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r, np.pi / 5)

        pq.Q(0, 1) | pq.Beamsplitter()

    simulator = pq.PureFockSimulator(d=d, config=pq.Config(cutoff=cutoff))
    state = simulator.execute(program).state

    variance = state.variance_photon_number()

    assert np.isclose(variance, r**2)


def test_PureFockState_unnormalized_state_raises_InvalidState():
    d = 1
    low_cutoff = 3

    high_displacement = 3.0

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(high_displacement)

    simulator = pq.PureFockSimulator(d=d, config=pq.Config(cutoff=low_cutoff))
    state = simulator.execute(program).state

    with pytest.raises(pq.api.exceptions.InvalidState) as error:
        state.validate()

    assert "The sum of probabilities is" in error.value.args[0]


def test_PureFockState_unnormalized_state_validate_False():
    d = 1
    low_cutoff = 3

    high_displacement = 3.0

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(high_displacement)

    simulator = pq.PureFockSimulator(
        d=d, config=pq.Config(cutoff=low_cutoff, validate=False)
    )
    state = simulator.execute(program).state

    assert not np.isclose(state.norm, 1.0)

    state.validate()



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import numpy as np
import piquasso as pq

import collections

from pytest_lazy_fixtures import lf

from scipy.stats import unitary_group
from scipy.linalg import polar, sinhm, coshm, expm


def is_proportional(first, second, rtol=1e-5):
    first = np.array(first)
    second = np.array(second)

    index = np.argmax(first)

    proportion = first[index] / second[index]

    return np.allclose(first, proportion * second, rtol=rtol)


tf_purefock_simulators = (
    lf("PureFockSimulator_with_tensorflow"),
    lf("PureFockSimulator_with_tensorflow_tf_function"),
)

jax_purefock_simulator = [
    lf("PureFockSimulator_with_jax"),
]


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.GaussianSimulator,
        pq.PureFockSimulator,
        *tf_purefock_simulators,
        pq.FockSimulator,
    ),
)
def test_fock_probabilities_should_be_numpy_array_of_floats(SimulatorClass):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(r=0.1, phi=0.6)

    simulator = SimulatorClass(d=3)
    state = simulator.execute(program).state

    probabilities = state.fock_probabilities

    assert isinstance(probabilities, collections.abc.Iterable)
    assert probabilities.dtype == np.float64


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.GaussianSimulator,
        pq.PureFockSimulator,
        *tf_purefock_simulators,
        *jax_purefock_simulator,
        pq.FockSimulator,
    ),
)
def test_fock_probabilities_with_squeezed_state(SimulatorClass):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(r=0.1, phi=0.6)

    simulator = SimulatorClass(d=3, config=pq.Config(cutoff=4))
    state = simulator.execute(program).state

    probabilities = state.fock_probabilities

    assert all(probability >= 0 for probability in probabilities)
    assert sum(probabilities) <= 1.0 or np.isclose(sum(probabilities), 1.0)

    assert is_proportional(
        probabilities,
        [
            0.99502075,
            0.0,
            0.0,
            0.0,
            0.00494212,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
    )


def test_density_matrix_with_squeezed_state():
    d = 2

    with pq.Program() as gaussian_program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(r=0.1, phi=np.pi / 3)

    gaussian_simulator = pq.GaussianSimulator(d=d)
    gaussian_state = gaussian_simulator.execute(gaussian_program).state

    gaussian_density_matrix = gaussian_state.density_matrix

    with pq.Program() as fock_program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(r=0.1, phi=np.pi / 3)

    fock_simulator = pq.FockSimulator(d=d)
    fock_state = fock_simulator.execute(fock_program).state

    fock_density_matrix = fock_state.density_matrix

    assert np.allclose(gaussian_density_matrix, fock_density_matrix)


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.GaussianSimulator,
        pq.PureFockSimulator,
        *tf_purefock_simulators,
        *jax_purefock_simulator,
        pq.FockSimulator,
    ),
)
def test_fock_probabilities_with_displaced_state(SimulatorClass):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=np.sqrt(5), phi=np.angle(1 + 2j))

    simulator = SimulatorClass(d=3)
    state = simulator.execute(program).state

    probabilities = state.fock_probabilities

    assert all(probability >= 0 for probability in probabilities)
    assert sum(probabilities) <= 1.0 or np.isclose(sum(probabilities), 1.0)

    assert is_proportional(
        probabilities,
        [
            0.00673795,
            0.03368973,
            0.0,
            0.0,
            0.08422434,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.1403739,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
    )


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.PureFockSimulator,
        pq.FockSimulator,
        *tf_purefock_simulators,
        *jax_purefock_simulator,
        pq.GaussianSimulator,
    ),
)
def test_Displacement_equivalence_on_multiple_modes(SimulatorClass):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=0.01)
        pq.Q(1) | pq.Displacement(r=0.02)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 4)

    simulator = SimulatorClass(d=2)

    state = simulator.execute(program).state

    assert is_proportional(
        state.fock_probabilities,
        [
            0.99950012,
            0.00020242,
            0.00029733,
            0.00000002,
            0.00000006,
            0.00000004,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
    )


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.GaussianSimulator,
        pq.PureFockSimulator,
        *tf_purefock_simulators,
        *jax_purefock_simulator,
        pq.FockSimulator,
    ),
)
def test_fock_probabilities_with_displaced_state_with_beamsplitter(SimulatorClass):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=np.sqrt(5), phi=np.angle(1 + 2j))
        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 4)

    simulator = SimulatorClass(d=3)
    state = simulator.execute(program).state

    probabilities = state.fock_probabilities

    assert all(probability >= 0 for probability in probabilities)
    assert sum(probabilities) <= 1.0 or np.isclose(sum(probabilities), 1.0)

    assert is_proportional(
        probabilities,
        [
            0.00673795,
            0.00842243,
            0.0252673,
            0.0,
            0.00526402,
            0.03158413,
            0.0,
            0.04737619,
            0.0,
            0.0,
            0.00219334,
            0.01974008,
            0.0,
            0.05922024,
            0.0,
            0.0,
            0.05922024,
            0.0,
            0.0,
            0.0,
        ],
    )


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.GaussianSimulator,
        pq.PureFockSimulator,
        *tf_purefock_simulators,
        *jax_purefock_simulator,
        pq.FockSimulator,
    ),
)
def test_fock_probabilities_with_squeezed_state_with_beamsplitter(SimulatorClass):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(r=0.1, phi=0.6)
        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 4)

    simulator = SimulatorClass(d=3)
    state = simulator.execute(program).state

    probabilities = state.fock_probabilities

    assert all(probability >= 0 for probability in probabilities)
    assert sum(probabilities) <= 1.0 or np.isclose(sum(probabilities), 1.0)

    assert is_proportional(
        probabilities,
        [
            0.99502075,
            0.0,
            0.0,
            0.0,
            0.00030888,
            0.0018533,
            0.0,
            0.00277994,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
    )


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.GaussianSimulator,
        pq.PureFockSimulator,
        *tf_purefock_simulators,
        *jax_purefock_simulator,
        pq.FockSimulator,
    ),
)
def test_fock_probabilities_with_two_single_mode_squeezings(SimulatorClass):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(r=0.1, phi=0.6)
        pq.Q(1) | pq.Squeezing(r=0.2, phi=0.7)

    simulator = SimulatorClass(d=2)
    state = simulator.execute(program).state

    probabilities = state.fock_probabilities

    assert all(probability >= 0 for probability in probabilities)
    assert sum(probabilities) <= 1.0 or np.isclose(sum(probabilities), 1.0)

    assert is_proportional(
        probabilities,
        [0.9754467, 0.0, 0.0, 0.0048449, 0.0, 0.01900025, 0.0, 0.0, 0.0, 0.0],
    )


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.GaussianSimulator,
        pq.PureFockSimulator,
        *tf_purefock_simulators,
        *jax_purefock_simulator,
        pq.FockSimulator,
    ),
)
def test_Squeezing_equivalence_on_multiple_modes(SimulatorClass):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(r=0.1, phi=np.pi / 5)
        pq.Q(1) | pq.Squeezing(r=0.2, phi=np.pi / 3)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 4)

    simulator = SimulatorClass(d=2)

    state = simulator.execute(program).state

    assert is_proportional(
        state.fock_probabilities,
        [0.97613795, 0.0, 0.0, 0.01246268, 0.00601937, 0.00537999, 0.0, 0.0, 0.0, 0.0],
    )


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.GaussianSimulator,
        pq.PureFockSimulator,
        *tf_purefock_simulators,
        *jax_purefock_simulator,
        pq.FockSimulator,
    ),
)
def test_fock_probabilities_with_two_mode_squeezing(SimulatorClass):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0, 1) | pq.Squeezing2(r=0.1, phi=0.6)

    simulator = SimulatorClass(d=3)
    state = simulator.execute(program).state

    probabilities = state.fock_probabilities

    # NOTE: It should not happen that the probabilities are negative, but in some cases
    # (possibly due to floating point errors) they turn out to be slightly negative.
    assert all(
        probability >= 0.0 or np.isclose(probability, 0.0)
        for probability in probabilities
    )
    assert sum(probabilities) <= 1.0 or np.isclose(sum(probabilities), 1.0)

    assert is_proportional(
        probabilities,
        [
            0.99006629,
            0.0,
            0.0,
            0.0,
            0.0,
            0.00983503,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
    )


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.GaussianSimulator,
        pq.PureFockSimulator,
        *tf_purefock_simulators,
        *jax_purefock_simulator,
        pq.FockSimulator,
    ),
)
def test_fock_probabilities_with_two_mode_squeezing_and_beamsplitter(SimulatorClass):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0, 1) | pq.Squeezing2(r=0.1, phi=0.6)
        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 4)

    simulator = SimulatorClass(d=3)
    state = simulator.execute(program).state

    probabilities = state.fock_probabilities

    assert all(probability >= 0 for probability in probabilities)
    assert sum(probabilities) <= 1.0 or np.isclose(sum(probabilities), 1.0)

    assert is_proportional(
        probabilities,
        [
            0.99006629,
            0.0,
            0.0,
            0.0,
            0.00368814,
            0.00245876,
            0.0,
            0.00368814,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
    )


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.GaussianSimulator,
        pq.PureFockSimulator,
        *tf_purefock_simulators,
        *jax_purefock_simulator,
        pq.FockSimulator,
    ),
)
def test_fock_probabilities_with_quadratic_phase(SimulatorClass):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.QuadraticPhase(s=0.4)

    simulator = SimulatorClass(d=3)
    state = simulator.execute(program).state

    probabilities = state.fock_probabilities
    expected_probabilities = [
        0.98058068,
        0.0,
        0.0,
        0.0,
        0.01885732,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
    ]

    assert all(probability >= 0 for probability in probabilities)
    assert sum(probabilities) <= 1.0 or np.isclose(sum(probabilities), 1.0)

    assert is_proportional(probabilities, expected_probabilities)


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.GaussianSimulator,
        pq.PureFockSimulator,
        *tf_purefock_simulators,
        *jax_purefock_simulator,
        pq.FockSimulator,
    ),
)
def test_fock_probabilities_with_position_displacement(SimulatorClass):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.PositionDisplacement(x=0.2)

    simulator = SimulatorClass(d=3)
    state = simulator.execute(program).state

    probabilities = state.fock_probabilities
    expected_probabilities = [
        0.96078944,
        0.03843158,
        0.0,
        0.0,
        0.00076863,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.00001025,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
    ]

    assert all(probability >= 0 for probability in probabilities)
    assert sum(probabilities) <= 1.0 or np.isclose(sum(probabilities), 1.0)

    assert is_proportional(probabilities, expected_probabilities)


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.GaussianSimulator,
        pq.PureFockSimulator,
        *tf_purefock_simulators,
        *jax_purefock_simulator,
        pq.FockSimulator,
    ),
)
def test_fock_probabilities_with_momentum_displacement(SimulatorClass):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.MomentumDisplacement(p=0.2)

    simulator = SimulatorClass(d=3)
    state = simulator.execute(program).state

    probabilities = state.fock_probabilities
    expected_probabilities = [
        0.96078944,
        0.03843158,
        0.0,
        0.0,
        0.00076863,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.00001025,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
    ]

    assert all(probability >= 0 for probability in probabilities)
    assert sum(probabilities) <= 1.0 or np.isclose(sum(probabilities), 1.0)

    assert is_proportional(probabilities, expected_probabilities)


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.GaussianSimulator,
        pq.PureFockSimulator,
        *tf_purefock_simulators,
        *jax_purefock_simulator,
        pq.FockSimulator,
    ),
)
def test_fock_probabilities_with_position_displacement_is_HBAR_independent(
    SimulatorClass,
):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.PositionDisplacement(x=0.4)

    simulator1 = SimulatorClass(d=3, config=pq.Config(hbar=2))
    simulator2 = SimulatorClass(d=3, config=pq.Config(hbar=42))

    state1 = simulator1.execute(program).state
    state2 = simulator2.execute(program).state

    probabilities1 = state1.fock_probabilities
    probabilities2 = state2.fock_probabilities

    assert np.allclose(probabilities1, probabilities2)


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.GaussianSimulator,
        pq.PureFockSimulator,
        *tf_purefock_simulators,
        *jax_purefock_simulator,
        pq.FockSimulator,
    ),
)
def test_fock_probabilities_with_momentum_displacement_is_HBAR_independent(
    SimulatorClass,
):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.MomentumDisplacement(p=0.4)

    simulator1 = SimulatorClass(d=3, config=pq.Config(hbar=2))
    simulator2 = SimulatorClass(d=3, config=pq.Config(hbar=42))

    state1 = simulator1.execute(program).state
    state2 = simulator2.execute(program).state

    probabilities1 = state1.fock_probabilities
    probabilities2 = state2.fock_probabilities

    assert np.allclose(probabilities1, probabilities2)


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.GaussianSimulator,
        pq.PureFockSimulator,
        *tf_purefock_simulators,
        *jax_purefock_simulator,
        pq.FockSimulator,
    ),
)
def test_fock_probabilities_with_general_gaussian_transform(SimulatorClass):
    squeezing_matrix = np.array(
        [
            [0.1, 0.2 + 0.3j],
            [0.2 + 0.3j, 0.1],
        ],
        dtype=complex,
    )

    rotation_matrix = np.array(
        [
            [1, 3 - 2j],
            [3 + 2j, 1],
        ],
        dtype=complex,
    )

    U, r = polar(squeezing_matrix)

    passive = expm(-1j * rotation_matrix) @ coshm(r)
    active = expm(-1j * rotation_matrix) @ U @ sinhm(r)

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0, 1) | pq.GaussianTransform(passive=passive, active=active)

    simulator = SimulatorClass(d=3)
    state = simulator.execute(program).state

    probabilities = state.fock_probabilities
    expected_probabilities = [
        0.88726976,
        0.0,
        0.0,
        0.0,
        0.03892269,
        0.02174349,
        0.0,
        0.05206405,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
    ]

    assert all(probability >= 0 for probability in probabilities)
    assert sum(probabilities) <= 1.0 or np.isclose(sum(probabilities), 1.0)

    assert is_proportional(probabilities, expected_probabilities, rtol=1e-4)


@pytest.mark.monkey
def test_monkey_fock_probabilities_with_general_gaussian_transform(
    generate_unitary_matrix, generate_complex_symmetric_matrix
):
    d = 3

    squeezing_matrix = generate_complex_symmetric_matrix(3)
    U, r = polar(squeezing_matrix)

    global_phase = generate_unitary_matrix(d)
    passive = global_phase @ coshm(r)
    active = global_phase @ sinhm(r) @ U.conj()

    with pq.Program() as fock_program:
        pq.Q() | pq.Vacuum()

        pq.Q(all) | pq.GaussianTransform(passive=passive, active=active)

    fock_simulator = pq.FockSimulator(d=d)
    fock_state = fock_simulator.execute(fock_program).state

    fock_representation_probabilities = fock_state.fock_probabilities

    with pq.Program() as gaussian_program:
        pq.Q() | pq.Vacuum()

        pq.Q(all) | pq.GaussianTransform(passive=passive, active=active)

    gaussian_simulator = pq.GaussianSimulator(d=d)
    gaussian_state = gaussian_simulator.execute(gaussian_program).state

    gaussian_representation_probabilities = gaussian_state.fock_probabilities

    assert np.allclose(
        fock_representation_probabilities,
        gaussian_representation_probabilities,
    )


@pytest.mark.monkey
def test_monkey_get_density_matrix_with_general_gaussian_transform(
    generate_unitary_matrix, generate_complex_symmetric_matrix
):
    d = 3

    squeezing_matrix = generate_complex_symmetric_matrix(3)
    U, r = polar(squeezing_matrix)

    global_phase = generate_unitary_matrix(d)
    passive = global_phase @ coshm(r)
    active = global_phase @ sinhm(r) @ U.conj()

    with pq.Program() as fock_program:
        pq.Q() | pq.Vacuum()

        pq.Q(all) | pq.GaussianTransform(passive=passive, active=active)

    fock_simulator = pq.FockSimulator(d=d)
    fock_state = fock_simulator.execute(fock_program).state

    fock_representation_probabilities = fock_state.fock_probabilities

    with pq.Program() as gaussian_program:
        pq.Q() | pq.Vacuum()

        pq.Q(all) | pq.GaussianTransform(passive=passive, active=active)

    gaussian_simulator = pq.GaussianSimulator(d=d)
    gaussian_state = gaussian_simulator.execute(gaussian_program).state

    gaussian_representation_probabilities = gaussian_state.fock_probabilities

    assert np.allclose(
        fock_representation_probabilities,
        gaussian_representation_probabilities,
    )


def test_sampling_backend_equivalence_for_two_mode_beamsplitter():
    initial_occupation_numbers = (1, 1)
    d = len(initial_occupation_numbers)

    with pq.Program() as program:
        pq.Q() | pq.StateVector(initial_occupation_numbers)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 4)

    config = pq.Config(cutoff=sum(initial_occupation_numbers) + 1)

    fock_simulator = pq.PureFockSimulator(d=d, config=config)
    fock_state = fock_simulator.execute(program).state
    fock_state.validate()

    sampling_simulator = pq.SamplingSimulator(d=d, config=config)
    sampling_state = sampling_simulator.execute(program).state
    sampling_state.validate()

    assert np.allclose(
        fock_state.fock_probabilities,
        sampling_state.fock_probabilities,
    )


def test_sampling_backend_equivalence_two_mode_Beamsplitter5050_fock_probabilities():
    initial_occupation_numbers = (1, 1)
    d = len(initial_occupation_numbers)

    with pq.Program() as program:
        pq.Q() | pq.StateVector(initial_occupation_numbers)

        pq.Q(0, 1) | pq.Beamsplitter5050()

    config = pq.Config(cutoff=sum(initial_occupation_numbers) + 1)

    fock_simulator = pq.PureFockSimulator(d=d, config=config)
    fock_state = fock_simulator.execute(program).state
    fock_state.validate()

    sampling_simulator = pq.SamplingSimulator(d=d, config=config)
    sampling_state = sampling_simulator.execute(program).state
    sampling_state.validate()

    assert np.allclose(
        fock_state.fock_probabilities,
        sampling_state.fock_probabilities,
    )


def test_sampling_backend_equivalence_two_mode_Beamsplitter5050_state_vector():
    initial_occupation_numbers = (1, 1)
    d = len(initial_occupation_numbers)

    with pq.Program() as program:
        pq.Q() | pq.StateVector(initial_occupation_numbers)

        pq.Q(0, 1) | pq.Beamsplitter5050()

    config = pq.Config(cutoff=sum(initial_occupation_numbers) + 1)

    fock_simulator = pq.PureFockSimulator(d=d, config=config)
    fock_state = fock_simulator.execute(program).state
    fock_state.validate()

    sampling_simulator = pq.SamplingSimulator(d=d, config=config)
    sampling_state = sampling_simulator.execute(program).state
    sampling_state.validate()

    assert np.allclose(
        fock_state.state_vector,
        sampling_state.state_vector,
    )


def test_sampling_backend_equivalence_complex_scenario_fock_probabilities():
    initial_occupation_numbers = (1, 1, 0, 1)
    d = len(initial_occupation_numbers)

    with pq.Program() as program:
        pq.Q() | pq.StateVector(initial_occupation_numbers)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 4)

        pq.Q(1) | pq.Phaseshifter(np.pi / 3)

        pq.Q(1, 2) | pq.Beamsplitter5050()

    config = pq.Config(cutoff=sum(initial_occupation_numbers) + 1)

    fock_simulator = pq.PureFockSimulator(d=d, config=config)
    fock_state = fock_simulator.execute(program).state
    fock_state.validate()

    sampling_simulator = pq.SamplingSimulator(d=d, config=config)
    sampling_state = sampling_simulator.execute(program).state
    sampling_state.validate()

    assert np.allclose(fock_state.fock_probabilities, sampling_state.fock_probabilities)


def test_sampling_backend_equivalence_complex_scenario_state_vector():
    initial_occupation_numbers = (1, 1, 0, 1)
    d = len(initial_occupation_numbers)

    with pq.Program() as program:
        pq.Q() | pq.StateVector(initial_occupation_numbers)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 4)

        pq.Q(1) | pq.Phaseshifter(np.pi / 3)

        pq.Q(1, 2) | pq.Beamsplitter5050()

    config = pq.Config(cutoff=sum(initial_occupation_numbers) + 1)

    fock_simulator = pq.PureFockSimulator(d=d, config=config)
    fock_state = fock_simulator.execute(program).state
    fock_state.validate()

    sampling_simulator = pq.SamplingSimulator(d=d, config=config)
    sampling_state = sampling_simulator.execute(program).state
    sampling_state.validate()

    assert np.allclose(fock_state.state_vector, sampling_state.state_vector)


def test_sampling_backend_equivalence_complex_scenario_fock_probabilities_2():
    initial_occupation_numbers = np.array(
        [
            [1, 1, 0, 1],
            [2, 0, 0, 1],
            [1, 1, 0, 0],
            [0, 0, 0, 0],
        ]
    )
    coefficients = np.sqrt([0.1, 0.2, 0.3, 0.4])
    d = initial_occupation_numbers.shape[1]

    with pq.Program() as program:
        for idx in range(len(coefficients)):
            pq.Q() | pq.StateVector(
                initial_occupation_numbers[idx], coefficient=coefficients[idx]
            )

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 4)

        pq.Q(1) | pq.Phaseshifter(np.pi / 3)

        pq.Q(1, 2) | pq.Beamsplitter5050()

    config = pq.Config(cutoff=4)

    fock_simulator = pq.PureFockSimulator(d=d, config=config)
    fock_state = fock_simulator.execute(program).state
    fock_state.validate()

    sampling_simulator = pq.SamplingSimulator(d=d, config=config)
    sampling_state = sampling_simulator.execute(program).state
    sampling_state.validate()

    assert np.allclose(fock_state.fock_probabilities, sampling_state.fock_probabilities)


def test_sampling_backend_equivalence_complex_scenario_state_vector_2():
    initial_occupation_numbers = np.array(
        [
            [1, 1, 0, 1],
            [2, 0, 0, 1],
            [1, 1, 0, 0],
            [0, 0, 0, 0],
        ]
    )
    coefficients = np.sqrt([0.1, 0.2, 0.3, 0.4])
    d = initial_occupation_numbers.shape[1]

    with pq.Program() as program:
        for idx in range(len(coefficients)):
            pq.Q() | pq.StateVector(
                initial_occupation_numbers[idx], coefficient=coefficients[idx]
            )

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 4)

        pq.Q(1) | pq.Phaseshifter(np.pi / 3)

        pq.Q(1, 2) | pq.Beamsplitter5050()

    config = pq.Config(cutoff=4)

    fock_simulator = pq.PureFockSimulator(d=d, config=config)
    fock_state = fock_simulator.execute(program).state
    fock_state.validate()

    sampling_simulator = pq.SamplingSimulator(d=d, config=config)
    sampling_state = sampling_simulator.execute(program).state
    sampling_state.validate()

    assert np.allclose(fock_state.state_vector, sampling_state.state_vector)


@pytest.mark.monkey
def test_sampling_backend_equivalence_with_random_interferometer_fock_probabilities(
    generate_unitary_matrix,
):
    initial_occupation_numbers = (1, 1, 0, 1)
    d = len(initial_occupation_numbers)

    interferometer_matrix = generate_unitary_matrix(d)

    with pq.Program() as program:
        pq.Q() | pq.StateVector(initial_occupation_numbers)

        pq.Q(all) | pq.Interferometer(interferometer_matrix)

    config = pq.Config(cutoff=sum(initial_occupation_numbers) + 1)

    fock_simulator = pq.PureFockSimulator(d=d, config=config)
    fock_state = fock_simulator.execute(program).state
    fock_state.validate()

    sampling_simulator = pq.SamplingSimulator(d=d, config=config)
    sampling_state = sampling_simulator.execute(program).state
    sampling_state.validate()

    assert np.allclose(
        fock_state.fock_probabilities,
        sampling_state.fock_probabilities,
    )


@pytest.mark.monkey
def test_sampling_backend_equivalence_with_random_interferometer_state_vector(
    generate_unitary_matrix,
):
    initial_occupation_numbers = (1, 1, 0, 1)
    d = len(initial_occupation_numbers)

    interferometer_matrix = generate_unitary_matrix(d)

    with pq.Program() as program:
        pq.Q() | pq.StateVector(initial_occupation_numbers)

        pq.Q(all) | pq.Interferometer(interferometer_matrix)

    config = pq.Config(cutoff=sum(initial_occupation_numbers) + 1)

    fock_simulator = pq.PureFockSimulator(d=d, config=config)
    fock_state = fock_simulator.execute(program).state
    fock_state.validate()

    sampling_simulator = pq.SamplingSimulator(d=d, config=config)
    sampling_state = sampling_simulator.execute(program).state
    sampling_state.validate()

    assert np.allclose(
        fock_state.state_vector,
        sampling_state.state_vector,
    )


def test_wigner_function_equivalence():
    config = pq.Config(cutoff=10, hbar=42)

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=0.10, phi=np.angle(1 - 0.5j))
        pq.Q(0) | pq.Squeezing(r=0.10)

    fock_simulator = pq.FockSimulator(d=1, config=config)
    fock_state = fock_simulator.execute(program).state

    fock_wigner_function_values = fock_state.wigner_function(
        positions=[0.10, 0.11],
        momentums=[-0.05, -0.06],
    )

    gaussian_simulator = pq.GaussianSimulator(d=1, config=config)
    gaussian_state = gaussian_simulator.execute(program).state

    gaussian_wigner_function_values = gaussian_state.wigner_function(
        positions=[[0.10], [0.11]],
        momentums=[[-0.05], [-0.06]],
    )

    assert np.allclose(fock_wigner_function_values, gaussian_wigner_function_values)


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.GaussianSimulator,
        pq.PureFockSimulator,
        pq.FockSimulator,
    ),
)
def test_fidelity_for_1_mode(SimulatorClass):
    with pq.Program() as program_1:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Squeezing(r=0.2, phi=-np.pi / 3)
        pq.Q(0) | pq.Displacement(r=-0.1, phi=0)

    with pq.Program() as program_2:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Squeezing(r=0.2, phi=np.pi / 3)
        pq.Q(0) | pq.Displacement(r=-0.1, phi=0)

    generic_simulator = SimulatorClass(d=1, config=pq.Config(cutoff=10))

    state_1 = generic_simulator.execute(program_1).state
    state_2 = generic_simulator.execute(program_2).state
    fidelity = state_1.fidelity(state_2)

    assert np.isclose(fidelity, 0.9421652615828)
    assert np.isclose(fidelity, state_2.fidelity(state_1))


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.GaussianSimulator,
        pq.PureFockSimulator,
        pq.FockSimulator,
    ),
)
def test_fidelity_for_nondisplaced_states_on_2_modes(SimulatorClass):
    with pq.Program() as program_1:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(0.1)
        pq.Q(1) | pq.Squeezing(0.2)

        pq.Q(all) | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 7)

    with pq.Program() as program_2:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(0.3)
        pq.Q(1) | pq.Squeezing(0.1)

        pq.Q(all) | pq.Beamsplitter(theta=np.pi / 4, phi=np.pi / 9)

    simulator = SimulatorClass(d=2, config=pq.Config(cutoff=9))

    state_1 = simulator.execute(program_1).state
    state_2 = simulator.execute(program_2).state
    fidelity = state_1.fidelity(state_2)

    assert np.isclose(fidelity, 0.9735085877314046)
    assert np.isclose(fidelity, state_2.fidelity(state_1))


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.GaussianSimulator,
        pq.PureFockSimulator,
        pq.FockSimulator,
    ),
)
def test_fidelity_for_displaced_states_on_2_modes(SimulatorClass):
    with pq.Program() as program_1:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(0.4)
        pq.Q(1) | pq.Displacement(0.5)

        pq.Q(0) | pq.Squeezing(0.1)
        pq.Q(1) | pq.Squeezing(0.2)

        pq.Q(all) | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 7)

    with pq.Program() as program_2:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(0.5)
        pq.Q(1) | pq.Displacement(0.4)

        pq.Q(0) | pq.Squeezing(0.3)
        pq.Q(1) | pq.Squeezing(0.1)

        pq.Q(all) | pq.Beamsplitter(theta=np.pi / 4, phi=np.pi / 9)

    simulator = SimulatorClass(d=2, config=pq.Config(cutoff=10))

    state_1 = simulator.execute(program_1).state
    state_2 = simulator.execute(program_2).state
    fidelity = state_1.fidelity(state_2)

    assert np.isclose(fidelity, 0.9346675071279842)
    assert np.isclose(fidelity, state_2.fidelity(state_1))


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.GaussianSimulator,
        pq.PureFockSimulator,
        pq.FockSimulator,
    ),
)
def test_fidelity_for_nondisplaced_pure_states_on_3_modes(SimulatorClass):
    with pq.Program() as program_1:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(0.04)
        pq.Q(1) | pq.Displacement(0.05)
        pq.Q(2) | pq.Displacement(0.1)

        pq.Q(0) | pq.Squeezing(0.01)
        pq.Q(1) | pq.Squeezing(0.02)
        pq.Q(2) | pq.Squeezing(0.03)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 7)
        pq.Q(1, 2) | pq.Beamsplitter(theta=np.pi / 4, phi=np.pi / 9)

    with pq.Program() as program_2:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(0.05)
        pq.Q(1) | pq.Displacement(0.04)
        pq.Q(2) | pq.Displacement(0.02)

        pq.Q(0) | pq.Squeezing(0.03)
        pq.Q(1) | pq.Squeezing(0.01)
        pq.Q(2) | pq.Squeezing(0.02)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 2, phi=np.pi / 9)
        pq.Q(1, 2) | pq.Beamsplitter(theta=np.pi / 5, phi=np.pi / 3)

    simulator = SimulatorClass(d=3, config=pq.Config(cutoff=7))

    state_1 = simulator.execute(program_1).state
    state_2 = simulator.execute(program_2).state
    fidelity = state_1.fidelity(state_2)

    assert np.isclose(fidelity, 0.9889124929545777)
    assert np.isclose(fidelity, state_2.fidelity(state_1))


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.GaussianSimulator,
        pq.PureFockSimulator,
        pq.FockSimulator,
    ),
)
def test_fidelity_for_nondisplaced_mixed_states_on_3_modes(SimulatorClass):
    with pq.Program() as program_1:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(0.04)
        pq.Q(1) | pq.Displacement(0.05)
        pq.Q(2) | pq.Displacement(0.1)

        pq.Q(0) | pq.Squeezing(0.01)
        pq.Q(1) | pq.Squeezing(0.02)
        pq.Q(2) | pq.Squeezing(0.03)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 7)
        pq.Q(1, 2) | pq.Beamsplitter(theta=np.pi / 4, phi=np.pi / 9)

    with pq.Program() as program_2:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(0.05)
        pq.Q(1) | pq.Displacement(0.04)
        pq.Q(2) | pq.Displacement(0.02)

        pq.Q(0) | pq.Squeezing(0.03)
        pq.Q(1) | pq.Squeezing(0.01)
        pq.Q(2) | pq.Squeezing(0.02)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 2, phi=np.pi / 9)
        pq.Q(1, 2) | pq.Beamsplitter(theta=np.pi / 5, phi=np.pi / 3)

    simulator = SimulatorClass(d=3, config=pq.Config(cutoff=7))

    state_1 = simulator.execute(program_1).state.reduced(modes=(0, 1))
    state_2 = simulator.execute(program_2).state.reduced(modes=(0, 1))
    fidelity = state_1.fidelity(state_2)

    assert np.isclose(fidelity, 0.9959871270027937)
    assert np.isclose(fidelity, state_2.fidelity(state_1))


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.PureFockSimulator,
        pq.FockSimulator,
        *tf_purefock_simulators,
    ),
)
def test_cubic_phase_equivalency(SimulatorClass):
    with pq.Program() as program_1:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.CubicPhase(gamma=0.1)
        pq.Q(1) | pq.CubicPhase(gamma=-0.07)

    with pq.Program() as program_2:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.CubicPhase(gamma=0.1)
        pq.Q(1) | pq.CubicPhase(gamma=-0.07)

    generic_simulator = SimulatorClass(d=2, config=pq.Config(cutoff=10))

    state_1 = generic_simulator.execute(program_1).state
    state_2 = generic_simulator.execute(program_2).state
    fidelity = state_1.fidelity(state_2)

    assert np.isclose(fidelity, 1.0)
    assert np.isclose(fidelity, state_2.fidelity(state_1))


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.PureFockSimulator,
        pq.FockSimulator,
        *tf_purefock_simulators,
    ),
)
def test_CubicPhase_equivalence_on_multiple_modes(SimulatorClass):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.CubicPhase(gamma=0.1)
        pq.Q(1) | pq.CubicPhase(gamma=0.2)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 4)

    simulator = SimulatorClass(d=2)

    state = simulator.execute(program).state

    assert is_proportional(
        state.fock_probabilities,
        [
            0.97960898,
            0.00474935,
            0.00710543,
            0.00029005,
            0.00030477,
            0.0001192,
            0.00306916,
            0.00174448,
            0.00261566,
            0.00039293,
        ],
    )


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.PureFockSimulator,
        pq.FockSimulator,
        *tf_purefock_simulators,
    ),
)
def test_Kerr_gate_leaves_fock_probabilities_invariant(SimulatorClass):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Squeezing(r=0.1, phi=np.pi / 5)
        pq.Q(1) | pq.Squeezing(r=0.2, phi=np.pi / 6)

        pq.Q(0) | pq.Kerr(xi=-0.1)
        pq.Q(1) | pq.Kerr(xi=0.2)

    simulator = SimulatorClass(d=2)

    state = simulator.execute(program).state

    assert is_proportional(
        state.fock_probabilities,
        [0.97613795, 0.0, 0.0, 0.00484834, 0.0, 0.01901371, 0.0, 0.0, 0.0, 0.0],
    )


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.PureFockSimulator,
        *tf_purefock_simulators,
    ),
)
def test_Kerr_equivalence(SimulatorClass):
    xi = np.pi / 4

    n = 2

    with pq.Program() as program:
        pq.Q(all) | pq.StateVector([0]) / np.sqrt(2)
        pq.Q(all) | pq.StateVector([n]) / np.sqrt(2)

        pq.Q(all) | pq.Kerr(xi=xi)

    simulator = SimulatorClass(d=1, config=pq.Config(cutoff=n + 1))

    state = simulator.execute(program).state

    eixi = np.exp(1j * xi * n * n) / 2
    emixi = np.exp(-1j * xi * n * n) / 2

    assert np.allclose(
        state.density_matrix, [[0.5, 0, emixi], [0, 0, 0], [eixi, 0, 0.5]]
    )


def test_Kerr_equivalence_for_FockSimulator():
    xi = np.pi / 4

    n = 2

    with pq.Program() as program:
        pq.Q(all) | pq.DensityMatrix(ket=(0,), bra=(0,)) / 2
        pq.Q(all) | pq.DensityMatrix(ket=(0,), bra=(2,)) / 2
        pq.Q(all) | pq.DensityMatrix(ket=(2,), bra=(0,)) / 2
        pq.Q(all) | pq.DensityMatrix(ket=(2,), bra=(2,)) / 2

        pq.Q(0) | pq.Kerr(xi=xi)

    simulator = pq.FockSimulator(d=1, config=pq.Config(cutoff=n + 1))

    state = simulator.execute(program).state

    eixi = np.exp(1j * xi * n * n) / 2
    emixi = np.exp(-1j * xi * n * n) / 2

    assert np.allclose(
        state.density_matrix, [[0.5, 0, emixi], [0, 0, 0], [eixi, 0, 0.5]]
    )


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.PureFockSimulator,
        pq.FockSimulator,
        pq.GaussianSimulator,
    ),
)
def test_Attenuator_equivalence_on_one_mode(SimulatorClass):
    config = pq.Config(cutoff=7)

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(r=0.1)

        pq.Q(0) | pq.Phaseshifter(phi=np.pi / 12)

        pq.Q(0) | pq.Attenuator(theta=np.pi / 3)

    simulator = SimulatorClass(d=1, config=config)

    state = simulator.execute(program).state

    assert is_proportional(
        state.fock_probabilities,
        [0.9978124, 0.00186894, 0.00031674, 0.00000177, 0.00000015, 0.0, 0.0],
    )


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.PureFockSimulator,
        pq.FockSimulator,
        pq.GaussianSimulator,
    ),
)
def test_Attenuator_equivalence(SimulatorClass):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=0.01)
        pq.Q(1) | pq.Displacement(r=0.02)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 4)

        pq.Q(0) | pq.Attenuator(theta=np.pi / 3)

    simulator = SimulatorClass(d=2)

    state = simulator.execute(program).state

    assert is_proportional(
        state.fock_probabilities,
        [
            0.99965195,
            0.00005061,
            0.00029737,
            0.0,
            0.00000002,
            0.00000004,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
    )


@pytest.mark.parametrize(
    "SimulatorClass", (pq.GaussianSimulator, pq.PureFockSimulator, pq.FockSimulator)
)
def test_Attenuator_and_Beamsplitter_with_ancilla_qumode_equivalence(SimulatorClass):
    theta = np.pi / 3

    with pq.Program() as lossy_program:
        pq.Q() | pq.Vacuum()

        pq.Q(0, 1) | pq.Squeezing2(0.03)

        pq.Q(0) | pq.Attenuator(theta=theta)

    with pq.Program() as program_with_ancilla_qumode:
        pq.Q() | pq.Vacuum()

        pq.Q(0, 1) | pq.Squeezing2(0.03)

        pq.Q(0, 2) | pq.Beamsplitter(theta=theta, phi=np.pi / 4)

    lossy_simulator = SimulatorClass(d=2)
    ancilla_simulator = SimulatorClass(d=3)

    lossy_state = lossy_simulator.execute(lossy_program).state

    state_with_ancilla_qumode = ancilla_simulator.execute(
        program_with_ancilla_qumode
    ).state
    reduced_state = state_with_ancilla_qumode.reduced(modes=(0, 1))

    assert reduced_state == lossy_state


@pytest.mark.parametrize(
    "SimulatorClass", (pq.GaussianSimulator, pq.PureFockSimulator, pq.FockSimulator)
)
def test_Attenuator_raises_InvalidInstruction_for_multiple_modes(SimulatorClass):
    theta = np.pi / 6
    mean_thermal_excitation = 5

    with pq.Program() as program:
        pq.Q(0, 1) | pq.Attenuator(
            theta=theta, mean_thermal_excitation=mean_thermal_excitation
        )

    simulator = SimulatorClass(d=2)

    with pytest.raises(pq.api.exceptions.InvalidInstruction) as error:
        simulator.execute(program)

    assert "The instruction should be specified for '2' modes:" in error.value.args[0]


@pytest.mark.parametrize("SimulatorClass", (pq.PureFockSimulator, pq.FockSimulator))
def test_Attenuator_raises_InvalidParam_for_non_zero_mean_thermal_excitation(
    SimulatorClass,
):
    theta = np.pi / 6
    nonzero_mean_thermal_excitation = 5

    with pq.Program() as program:
        pq.Q(0) | pq.Attenuator(
            theta=theta, mean_thermal_excitation=nonzero_mean_thermal_excitation
        )

    simulator = SimulatorClass(d=2)

    with pytest.raises(pq.api.exceptions.InvalidParameter) as error:
        simulator.execute(program)

    assert (
        "Non-zero mean thermal excitation is not supported in this backend."
        in error.value.args[0]
    )


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.PureFockSimulator,
        pq.FockSimulator,
        pq.GaussianSimulator,
        *tf_purefock_simulators,
    ),
)
def test_Interferometer_smaller_than_system_size(SimulatorClass):
    config = pq.Config(cutoff=3)

    d = 5

    interferometer_matrix = np.array(
        [
            [
                0.67622072 - 0.02632995j,
                -0.41993552 - 0.43404319j,
                0.21135489 + 0.36417311j,
            ],
            [
                0.48143126 - 0.18351759j,
                -0.05256469 + 0.2714796j,
                -0.7896096 - 0.18600457j,
            ],
            [
                0.47764293 - 0.22007895j,
                0.74644107 + 0.0402763j,
                0.35346034 - 0.19922808j,
            ],
        ]
    )

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        for i in range(d):
            pq.Q(i) | pq.Squeezing(r=0.01 * i)

        pq.Q(0, 4, 2) | pq.Interferometer(interferometer_matrix)

    simulator = SimulatorClass(d=5, config=config)

    state = simulator.execute(program).state

    assert is_proportional(
        state.fock_probabilities,
        [
            0.99850342,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.00015807,
            0.0,
            0.0002202,
            0.0,
            0.00011654,
            0.00004992,
            0.0,
            0.0,
            0.0,
            0.00028562,
            0.0,
            0.00016609,
            0.00044906,
            0.0,
            0.00005109,
        ],
    )


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.PureFockSimulator,
        pq.FockSimulator,
        pq.GaussianSimulator,
        *tf_purefock_simulators,
    ),
)
def test_Squeezing_with_zero_parameter(SimulatorClass):
    config = pq.Config(cutoff=3)

    d = 2

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(r=0.0)
        pq.Q(1) | pq.Squeezing(r=0.1)

    simulator = SimulatorClass(d=d, config=config)

    state = simulator.execute(program).state

    assert is_proportional(
        state.fock_probabilities,
        [0.99505769, 0.0, 0.0, 0.0, 0.0, 0.00494231],
    )


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.PureFockSimulator,
        pq.FockSimulator,
        pq.GaussianSimulator,
        *tf_purefock_simulators,
    ),
)
def test_Displacement_with_zero_parameter(SimulatorClass):
    config = pq.Config(cutoff=3)

    d = 2

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=0.0)
        pq.Q(1) | pq.Displacement(r=0.1)

    simulator = SimulatorClass(d=d, config=config)

    state = simulator.execute(program).state

    assert is_proportional(
        state.fock_probabilities,
        [0.99005, 0.0, 0.0099005, 0.0, 0.0, 0.0000495],
    )


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.PureFockSimulator,
        pq.FockSimulator,
        pq.GaussianSimulator,
        *tf_purefock_simulators,
    ),
)
def test_beamsplitter5050_equivalence(SimulatorClass):
    with pq.Program() as preparation:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=0.1, phi=np.pi / 5)
        pq.Q(1) | pq.Squeezing(r=0.1, phi=np.pi / 5)
        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 7, phi=np.pi / 11)

    with pq.Program() as program:
        pq.Q() | preparation

        pq.Q(0, 1) | pq.Beamsplitter5050()

    simulator = SimulatorClass(d=2, config=pq.Config(cutoff=5))
    state = simulator.execute(program).state

    assert np.allclose(
        state.fock_probabilities,
        [
            0.98512079,
            0.0012306,
            0.0086206,
            0.00364572,
            0.00128432,
            0.00001217,
            0.00001391,
            0.00001611,
            0.00001785,
            0.00000122,
            0.00002023,
            0.00001428,
            0.00000213,
            0.00000005,
            0.0,
        ],
    )


@pytest.mark.monkey
@pytest.mark.parametrize("SimulatorClass", (pq.PureFockSimulator, pq.SamplingSimulator))
def test_post_select_NS_gate(SimulatorClass):
    d = 3

    first_mode_state_vector = np.sqrt([0.2, 0.3, 0.5])

    ns_gate_interferometer = np.array(
        [
            [1 - 2 ** (1 / 2), 2 ** (-1 / 4), np.sqrt(3 / 2 ** (1 / 2) - 2)],
            [2 ** (-1 / 4), 1 / 2, 1 / 2 - np.sqrt(1 / 2)],
            [np.sqrt(3 / 2 ** (1 / 2) - 2), 1 / 2 - np.sqrt(1 / 2), np.sqrt(2) - 1 / 2],
        ]
    )

    with pq.Program() as program:
        pq.Q(all) | pq.StateVector([0, 1, 0]) * first_mode_state_vector[0]
        pq.Q(all) | pq.StateVector([1, 1, 0]) * first_mode_state_vector[1]
        pq.Q(all) | pq.StateVector([2, 1, 0]) * first_mode_state_vector[2]

        pq.Q(all) | pq.Interferometer(ns_gate_interferometer)

        pq.Q(all) | pq.PostSelectPhotons(postselect_modes=(1, 2), photon_counts=(1, 0))

    simulator = SimulatorClass(d=d, config=pq.Config(cutoff=4))

    state = simulator.execute(program).state

    assert np.isclose(state.norm, 0.25)

    assert np.isclose(state.norm, 0.25)

    state.normalize()
    state.validate()

    purity = state.get_purity()

    assert np.isclose(purity, 1.0)

    expected_state_vector = np.copy(first_mode_state_vector)
    expected_state_vector[2] *= -1

    assert np.allclose(
        state.density_matrix[:d, :d],
        np.outer(expected_state_vector, expected_state_vector),
    )


@pytest.mark.parametrize(
    "SimulatorClass",
    (pq.PureFockSimulator, pq.SamplingSimulator),
)
def test_post_select_random_unitary(SimulatorClass):
    d = 3

    interferometer_matrix = unitary_group.rvs(d)

    postselect_modes = (1, 2)

    with pq.Program() as program:
        pq.Q(all) | pq.StateVector([0, 1, 0]) * np.sqrt(0.2)
        pq.Q(all) | pq.StateVector([1, 1, 0]) * np.sqrt(0.3)
        pq.Q(all) | pq.StateVector([2, 1, 0]) * np.sqrt(0.5)

        pq.Q(all) | pq.Interferometer(interferometer_matrix)

        pq.Q(all) | pq.PostSelectPhotons(
            postselect_modes=postselect_modes, photon_counts=(1, 0)
        )

    simulator = SimulatorClass(d=d, config=pq.Config(cutoff=4))

    state = simulator.execute(program).state

    state.normalize()
    state.validate()

    assert state.d == d - len(postselect_modes)


@pytest.mark.parametrize(
    "SimulatorClass",
    (pq.PureFockSimulator, pq.SamplingSimulator),
)
def test_post_select_conditional_sign_flip_gate_with_1_over_16_success_rate(
    SimulatorClass,
):
    ancilla_modes = (4, 5, 6, 7)

    ns_gate_interferometer = np.array(
        [
            [1 - 2 ** (1 / 2), 2 ** (-1 / 4), np.sqrt(3 / 2 ** (1 / 2) - 2)],
            [2 ** (-1 / 4), 1 / 2, 1 / 2 - np.sqrt(1 / 2)],
            [np.sqrt(3 / 2 ** (1 / 2) - 2), 1 / 2 - np.sqrt(1 / 2), np.sqrt(2) - 1 / 2],
        ]
    )

    with pq.Program() as nonlinear_shift:
        pq.Q(all) | pq.Interferometer(ns_gate_interferometer)

    with pq.Program() as conditional_sign_flip:
        pq.Q(0, 2) | pq.Beamsplitter(theta=np.pi / 4)

        pq.Q(0, ancilla_modes[0], ancilla_modes[1]) | nonlinear_shift
        pq.Q(2, ancilla_modes[2], ancilla_modes[3]) | nonlinear_shift

        pq.Q(0, 2) | pq.Beamsplitter(theta=-np.pi / 4)

    state_00 = [0, 1, 0, 1]
    state_01 = [0, 1, 1, 0]
    state_10 = [1, 0, 0, 1]
    state_11 = [1, 0, 1, 0]

    states = [state_00, state_01, state_10, state_11]
    coefficients = np.sqrt([0.1, 0.2, 0.3, 0.4])

    ancilla_state = [1, 0]

    with pq.Program() as program:
        for input_state, coeff in zip(states, coefficients):
            pq.Q(all) | pq.StateVector(input_state + ancilla_state * 2) * coeff

        pq.Q(all) | conditional_sign_flip

        pq.Q(all) | pq.PostSelectPhotons(
            postselect_modes=ancilla_modes, photon_counts=ancilla_state * 2
        )

    simulator = SimulatorClass(d=8, config=pq.Config(cutoff=5))

    final_state = simulator.execute(program).state

    expected_success_rate = 1 / 16
    actual_success_rate = final_state.norm

    assert np.isclose(expected_success_rate, actual_success_rate)

    final_state.normalize()

    purity = final_state.get_purity()

    assert np.isclose(purity, 1.0)

    expected_coefficients = np.sqrt([0.1, 0.2, 0.3, 0.4])

    expected_coefficients[3] *= -1

    with pq.Program() as expectation_program:
        for input_state, coeff in zip(states, expected_coefficients):
            pq.Q(all) | pq.StateVector(input_state) * coeff

    expectation_simulator = SimulatorClass(d=4, config=pq.Config(cutoff=5))
    expected_state = expectation_simulator.execute(expectation_program).state

    assert np.allclose(expected_state.density_matrix, final_state.density_matrix)


@pytest.mark.parametrize(
    "SimulatorClass",
    (pq.PureFockSimulator, pq.SamplingSimulator),
)
def test_ImperfectPostSelectPhotons(SimulatorClass):
    d = 5
    cutoff = 4

    detector_efficiency_matrix = np.array(
        [
            [1.0, 0.2, 0.1],
            [0.0, 0.8, 0.2],
            [0.0, 0.0, 0.7],
        ]
    )

    coeffs = np.sqrt([0.1, 0.3, 0.4, 0.05, 0.1, 0.05])

    with pq.Program() as program:
        pq.Q() | pq.StateVector([0, 0, 0, 2, 1]) * coeffs[0]
        pq.Q() | pq.StateVector([0, 0, 2, 0, 1]) * coeffs[1]
        pq.Q() | pq.StateVector([0, 1, 0, 1, 1]) * coeffs[2]
        pq.Q() | pq.StateVector([1, 1, 0, 1, 0]) * coeffs[3]
        pq.Q() | pq.StateVector([3, 0, 0, 0, 0]) * coeffs[4]

        pq.Q() | pq.ImperfectPostSelectPhotons(
            postselect_modes=(2, 4),
            photon_counts=(0, 1),
            detector_efficiency_matrix=detector_efficiency_matrix,
        )

    simulator = SimulatorClass(d=d, config=pq.Config(cutoff=cutoff))

    state = simulator.execute(program).state

    with pq.Program() as expected_program:
        pq.Q() | pq.DensityMatrix((0, 1, 1), (0, 1, 1)) * 0.32
        pq.Q() | pq.DensityMatrix((0, 1, 1), (0, 0, 2)) * 0.16
        pq.Q() | pq.DensityMatrix((0, 0, 2), (0, 1, 1)) * 0.16
        pq.Q() | pq.DensityMatrix((0, 0, 2), (0, 0, 2)) * 0.08

    expected_simulator = pq.FockSimulator(d=3, config=pq.Config(cutoff=4))

    expected_state = expected_simulator.execute(expected_program).state

    assert state == expected_state


@pytest.mark.parametrize(
    "SimulatorClass",
    (pq.PureFockSimulator, pq.SamplingSimulator),
)
def test_NS_gate_with_ImperfectPostSelectPhotons_trivial_case(SimulatorClass):
    d = 3

    first_mode_state_vector = np.sqrt([0.2, 0.3, 0.5])

    ns_gate_interferometer = np.array(
        [
            [1 - 2 ** (1 / 2), 2 ** (-1 / 4), np.sqrt(3 / 2 ** (1 / 2) - 2)],
            [2 ** (-1 / 4), 1 / 2, 1 / 2 - np.sqrt(1 / 2)],
            [np.sqrt(3 / 2 ** (1 / 2) - 2), 1 / 2 - np.sqrt(1 / 2), np.sqrt(2) - 1 / 2],
        ]
    )

    trivial_detector_efficiency_matrix = np.identity(d)

    with pq.Program() as preparation:
        pq.Q(all) | pq.StateVector([0, 1, 0]) * first_mode_state_vector[0]
        pq.Q(all) | pq.StateVector([1, 1, 0]) * first_mode_state_vector[1]
        pq.Q(all) | pq.StateVector([2, 1, 0]) * first_mode_state_vector[2]

        pq.Q(all) | pq.Interferometer(ns_gate_interferometer)

    with pq.Program() as imperfect_photon_detection_program:
        pq.Q(all) | preparation

        pq.Q(all) | pq.ImperfectPostSelectPhotons(
            postselect_modes=(1, 2),
            photon_counts=(1, 0),
            detector_efficiency_matrix=trivial_detector_efficiency_matrix,
        )

    with pq.Program() as perfect_photon_detection_program:
        pq.Q(all) | preparation

        pq.Q(all) | pq.PostSelectPhotons(postselect_modes=(1, 2), photon_counts=(1, 0))

    simulator = SimulatorClass(d=d, config=pq.Config(cutoff=4))

    imperfect_state = simulator.execute(imperfect_photon_detection_program).state
    perfect_state = simulator.execute(perfect_photon_detection_program).state

    imperfect_success_rate = imperfect_state.norm
    perfect_success_rate = perfect_state.norm
    assert np.isclose(imperfect_success_rate, perfect_success_rate)
    assert np.isclose(perfect_success_rate, 1 / 4)

    imperfect_state.normalize()
    perfect_state.normalize()

    imperfect_state.validate()
    perfect_state.validate()

    assert np.isclose(imperfect_state.get_purity(), 1.0)

    assert np.allclose(imperfect_state.density_matrix, perfect_state.density_matrix)


@pytest.mark.parametrize(
    "SimulatorClass",
    (pq.PureFockSimulator, pq.SamplingSimulator),
)
def test_NS_gate_with_ImperfectPostSelectPhotons(SimulatorClass):
    d = 3

    first_mode_state_vector = np.sqrt([0.2, 0.3, 0.5])

    ns_gate_interferometer = np.array(
        [
            [1 - 2 ** (1 / 2), 2 ** (-1 / 4), np.sqrt(3 / 2 ** (1 / 2) - 2)],
            [2 ** (-1 / 4), 1 / 2, 1 / 2 - np.sqrt(1 / 2)],
            [np.sqrt(3 / 2 ** (1 / 2) - 2), 1 / 2 - np.sqrt(1 / 2), np.sqrt(2) - 1 / 2],
        ]
    )

    detector_efficiency_matrix = np.array(
        [[1.0, 0.1, 0.2], [0.0, 0.9, 0.2], [0.0, 0.0, 0.6]]
    )

    with pq.Program() as program:
        pq.Q(all) | pq.StateVector([0, 1, 0]) * first_mode_state_vector[0]
        pq.Q(all) | pq.StateVector([1, 1, 0]) * first_mode_state_vector[1]
        pq.Q(all) | pq.StateVector([2, 1, 0]) * first_mode_state_vector[2]

        pq.Q(all) | pq.Interferometer(ns_gate_interferometer)

        pq.Q(all) | pq.ImperfectPostSelectPhotons(
            postselect_modes=(1, 2),
            photon_counts=(1, 0),
            detector_efficiency_matrix=detector_efficiency_matrix,
        )

    simulator = SimulatorClass(d=d, config=pq.Config(cutoff=4))

    state = simulator.execute(program).state

    success_rate = state.norm

    assert np.isclose(success_rate, 0.25773863607376246)

    state.normalize()
    state.validate()

    assert np.isclose(state.get_purity(), 0.8645456946151449)

    assert np.allclose(
        state.density_matrix,
        [
            [0.25690057, 0.25784735, -0.27605969, 0.0],
            [0.25784735, 0.30661073, -0.33810269, 0.0],
            [-0.27605969, -0.33810269, 0.43648869, 0.0],
            [0.0, 0.0, 0.0, 0.0],
        ],
    )


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.PureFockSimulator,
        pq.GaussianSimulator,
        *tf_purefock_simulators,
    ),
)
def test_mean_photon_number_equivalence(SimulatorClass):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=0.1, phi=np.pi / 5)
        pq.Q(1) | pq.Squeezing(r=0.1, phi=np.pi / 5)

        pq.Q(0) | pq.Phaseshifter(np.pi / 9)
        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 7, phi=np.pi / 11)

    simulator = SimulatorClass(d=2, config=pq.Config(cutoff=10))
    state = simulator.execute(program).state

    assert np.allclose(state.mean_photon_number(), 0.02003337780953792)


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.PureFockSimulator,
        pq.GaussianSimulator,
        *tf_purefock_simulators,
    ),
)
def test_variance_photon_number_equivalence_1_mode(SimulatorClass):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=0.1, phi=np.pi / 5)

        pq.Q(0) | pq.Phaseshifter(np.pi / 9)

    simulator = SimulatorClass(d=1, config=pq.Config(cutoff=10))
    state = simulator.execute(program).state

    assert np.allclose(state.variance_photon_number(), 0.01)


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.PureFockSimulator,
        pq.GaussianSimulator,
        *tf_purefock_simulators,
    ),
)
def test_variance_photon_number_equivalence(SimulatorClass):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=0.1, phi=np.pi / 5)
        pq.Q(1) | pq.Squeezing(r=0.1, phi=np.pi / 5)

        pq.Q(0) | pq.Phaseshifter(np.pi / 9)
        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 7, phi=np.pi / 11)

    simulator = SimulatorClass(d=2, config=pq.Config(cutoff=10))
    state = simulator.execute(program).state

    assert np.allclose(state.variance_photon_number(), 0.030268090551006963)
